import java.awt.*;
import java.util.*;

/*****************************************************************
* A class that models a set of 2D random heights, such as the side
* on terrain found in simple games like WORMS.
* <p>The constructor randomly generates a set of heights of the
* specified size (length). Methods allow the drawing of the terrain
* (as a Ploygon or set of lines), flattening of areas, addition of
* craters, etc.
* @author Spike Barlow
* @date 1-3/8/1999
*******************************************************************/
public class RandomTerrain {

	private double[]      heights;


/******************************************************************
* Generate the Random Terrain such that it starts at the specified
* height, is of the prescribed length, and has the specified
* steepness. If the upper of lower bound is exceeded the height is
* reset to that bound.
*******************************************************************/
RandomTerrain(double initialHeight, int len, double steepness,
		double lowerBound, double upperBound) {
	double        bias = 0.0;
	int           i;
	Random        generator = new Random();
	double        nextValue;
	heights = new double[len];
	heights[0]=initialHeight;

	for (i=1;i<len;i++) {
		nextValue = heights[i-1] + generator.nextDouble()*2*steepness
			-steepness+bias;
		if (nextValue>upperBound)
			heights[i] = upperBound;
		else if (nextValue<lowerBound)
			heights[i] = lowerBound;
		else
			heights[i] = nextValue;

		// A bias term to encorage movement in the same general direction
		if (i>3) {
			bias = (heights[i]-heights[i-4])/4.0;
			if (bias>steepness/2)
				bias = steepness/2;
			else if (bias<-steepness/2)
				bias = -steepness/2;
		}
	}
}


/******************************************************************
* Generate the Random Terrain such that it starts at the specified
* height, is of the prescribed length, and has the specified
* steepness. There are no bounds on how high or low the terrain may
* grow.
********************************************************************/
RandomTerrain(double initialHeight, int len, double steepness) {

	this(initialHeight,len,steepness,Double.MIN_VALUE,Double.MAX_VALUE);
}


/*******************************************************************
* Draw the terrain as a set of connected lines on the Graphics object
* passed. Starting co-ordinates specify the start of the terrain on
* the graphics object and the scaling factors complete the mapping
* from Terrain to Graphics co-ordinate space. Generally the drawTerrain
* method is preferred.
**********************************************************************/
public void drawLineTerrain(Graphics g, int initialX, int initialY,
				double scaleX, double scaleY) {
	int			previousX,
				previousY,
				currentX,
				currentY;
	int			i;

	previousX = initialX;
	previousY = initialY - (int)Math.round(heights[0]*scaleY);
	for (i=1;i<heights.length;i++) {
		currentX = initialX + (int)Math.round(scaleX*i);
		currentY = initialY - (int)Math.round(scaleY*heights[i]);
		g.drawLine(previousX,previousY,currentX,currentY);
		previousX = currentX;
		previousY = currentY;
	}
}



/**************************************************************
* Draw the terrain as a set of solid polygon on the Graphics object
* passed. Starting co-ordinates specify the start of the terrain on
* the graphics object and the scaling factors complete the mapping
* from Terrain to Graphics co-ordinate space. A Color is provided
* together with lower bounds on the Polygon to be drawn. The Graphics
* object's original colour is restored prior to termination.
******************************************************************/
public void drawTerrain(Graphics g, int initialX, int initialY,
				double scaleX, double scaleY, Color land,
				int lowerX, int lowerY) {
	int			currentX,
				currentY;
	int			i;

	Color oldColour = g.getColor();
	g.setColor(land);
	Polygon outline = new Polygon();
	outline.addPoint(initialX,initialY);
	for (i=0;i<heights.length;i++) {
		outline.addPoint(initialX + (int)Math.round(scaleX*i),
		initialY - (int)Math.round(scaleY*heights[i]));
	}
	outline.addPoint(lowerX,lowerY);
	outline.addPoint(initialX,lowerY);
	g.fillPolygon(outline);
	g.setColor(oldColour);
}


/******************************************************************
* Draws the terrain as a set of connected lines on the Graphics
* object passed. As for the similar method but in this case only
* a single scaling factor is provided for both dimensions.
********************************************************************/
public void drawLineTerrain(Graphics g, int initialX, int initialY,
                double scale) {

	drawLineTerrain(g,initialX,initialY,scale,scale);
}



/*******************************************************************
* Returns the height of the highest peak on the terrain.
********************************************************************/
public double getMaxHeight() {
	double		max;
	int			i;

	max = heights[0];
	for (i=1;i<heights.length;i++)
		if (heights[i]>max)
		max = heights[i];
	return max;
}



/********************************************************************
* Returns the horizontal (X) location that corresponds to the highest
* peak.
*********************************************************************/
public double getHighestPeakPosition() {
	double        pos = 0;
	double        max;
	int           i;

	max = heights[0];
	for (i=1;i<heights.length;i++)
		if (heights[i]>max) {
			pos = i;
			max = heights[i];
		}
	return pos;
}



/***********************************************************************
* Finds the depth of the deepest valley on the terrain.
************************************************************************/
public double getMinHeight() {
	double        min;
	int           i;

	min = heights[0];
	for (i=1;i<heights.length;i++)
		if (heights[i]<min)
			min = heights[i];
	return min;
}



/*********************************************************************
* Returns the horizontal (X) location that corresponds to the deepest
* valley on the terrain.
**********************************************************************/
public double getLowestValleyPosition() {
	double        min;
	int           i;
	double        pos=0;

	min = heights[0];
	for (i=1;i<heights.length;i++)
		if (heights[i]<min) {
			pos = i;
			min = heights[i];
		}
	return pos;
}



/*******************************************************************
* returns the height at the specified horizontal location.
********************************************************************/
public double getHeightAt(int pos) {

	if (pos<0 || pos>=heights.length)
		return Double.MAX_VALUE;
	else
		return heights[pos];
}



/********************************************************************
* Returns the height at the specified horizontal position. A more
* sophisticated version of the method that allows any horizontal
* position rather than just an integer (array index) position. Uses
* linear interpolation to determine the height at the specified point.
**********************************************************************/
public double getHeightAt(double pos) {

	if (pos<0 || pos>=heights.length)
		return Double.MAX_VALUE;
	int upperIndex = (int)Math.ceil(pos);
	int lowerIndex = (int)Math.floor(pos);
	double weight = pos-lowerIndex;
	return (heights[lowerIndex]*(1.0-weight) + heights[upperIndex]*weight);
}


/**********************************************************************
* Add a crater to the terrain. A centre of the crater is specified plus
* an array specifying the depths. Each element of the array corresponds
* to the depth one "unit" further from the centre. The crater is assumed
* to be symetric: 0'th element is depth at centre, 1'st element is depth
* immediately on either side, etc.
************************************************************************/
public void addCrater(int pos, double[] depths) {
	double        centreDepth;
	int           i;

	if (pos<0 || pos>=heights.length)
		return;

	centreDepth = heights[pos];
	for (i=Math.max(0,pos-depths.length+1);
		i<Math.min(heights.length,pos+depths.length);i++)
		heights[i] = Math.min(centreDepth-depths[Math.abs(i-pos)],heights[i]);
}



/***********************************************************************
* Flatten (make level) a portion of the terrain. The central position of
* the flat region is supplied plus the total length of the flat area.
* All heights within the region are reset to be that found at the centre.
************************************************************************/
public void flatten(int pos, int length) {
	double        centreDepth;
	int           i;

	if (pos<0 || pos>=heights.length)
		return;

	centreDepth = heights[pos];
	for (i=Math.max(0,pos-length/2);
		i<Math.min(heights.length,pos+length/2);i++)
		heights[i] = centreDepth;
}

}
